-- CNO_NumpadUnitMovement
-- Author: MadManCam
-- DateCreated: 7/29/2018 4:26:02 PM
--------------------------------------------------------------

include( "Civ6Common" );
--include( "UnitPanel" );

local setting_KeyDownIgnoreTimer = 2.0;
local m_debug = false;

local m_bindings = {
	MOVE_SW = Keys.VK_NUMPAD1,
	MOVE_W	= Keys.VK_NUMPAD4,
	MOVE_NW = Keys.VK_NUMPAD7,
	MOVE_NE = Keys.VK_NUMPAD9,
	MOVE_E	= Keys.VK_NUMPAD6,
	MOVE_SE = Keys.VK_NUMPAD3,
	CENTER_CAMERA =	Keys.VK_NUMPAD5,
	SELECT_NEXT = Keys.VK_NUMPAD8,
	SELECT_PREV = Keys.VK_NUMPAD2
};
local m_double_bindings = {
	MOVE_SW = Keys.VK_NUMPAD1,
	MOVE_W	= Keys.VK_NUMPAD4,
	MOVE_NW = Keys.VK_NUMPAD7,
	MOVE_NE = Keys.VK_NUMPAD9,
	MOVE_E	= Keys.VK_NUMPAD6,
	MOVE_SE = Keys.VK_NUMPAD3,
	ATTACK_RANGE =	Keys.VK_NUMPAD5
};

local map_width, map_height = Map.GetGridSize();
local m_OkayToProcess = false;
local m_SelectedUnit = nil;
local m_InspectedEnemy = nil;
local m_IsSelectingRangeAttack = false;
local m_FocusedTargetPlot = -1;
local m_IsNumpadKeyDown = false;
local m_NumpadKey = -1;
local m_NumpadKeyTimePushed = -1;
local m_LastKeyUp = -1;
local m_LastKeyUpTime = -1;
local m_DoubleTapTimer = 1.0;
local m_UnitMoveIgnoredTimer = 2.0;

local UNIT_OPS = {};
for row in GameInfo.UnitOperations() do 
	local o = {};
	o.NAME = row.OperationType;
	o.HASH = row.Hash;
	UNIT_OPS[o.NAME] = o;
	if m_debug then	
		if ( o.NAME == "UNITOPERATION_MOVE_TO" ) or ( o.NAME == "UNITOPERATION_RANGE_ATTACK" ) then
			print( "Added Unit Operation: " .. o.NAME );
		end
	end
end

-- Get current system time with millisecond precision.
local function GetSystemTime()
	--[[
	local systime = -1;
	local a, b = math.modf( os.clock() );
	print( "a = " .. tostring( a ) .. "; b = " .. tostring( b ) );
	if b == 0 then b='000'; else b = tonumber( tostring( b ):sub( 3,5 ) ); end
	systime = a + b;
	return systime;
	--]]
	return os.clock();
end
if m_debug then print( "System time: " .. tostring( GetSystemTime() ) ); end

-- Unit move object
local m_UnitMoveClass = {
	ID = -1; -- unit ID
	owner = -1;	-- player id
	toPlot = -1; -- plot id
	timeStarted = -1; -- time started
	completed = false; -- completed successfully
	ignored = false; -- whether or not ignored.	
};
function m_UnitMoveClass:ignore() -- If the move should be ignored, tells unit to cancel and returns true.
	local timeSinceMoveStarted = GetSystemTime() - self.timeStarted;
	if timeSinceMoveStarted >= m_UnitMoveIgnoredTimer then -- Its past the timer, ignore this move and return false.
		self.ignored = true;
		local unit = UnitManager.GetUnit( self.owner, self.ID );
		if unit then
			UnitManager.RequestCommand( unit, UnitCommandTypes.CANCEL );	
		end
		return true;
	else
		return false;
	end
end
-- Object Constructor
function m_UnitMoveClass:new ( o )
	o = o or {};
	setmetatable( o, self );
	self.__index = self;
	return o;
end

-- Table of unit moves.
local m_UnitMoves = {
	members = {};
	index = 1;
	maxMoves = 30; -- Maximum number of moves saved in table.
};
function  m_UnitMoves:flush()
	self.members = {};
	self.index = 1;	
end
function  m_UnitMoves:add( o ) -- Adds a new object to the table.
	if o then 
		if self.index > self.maxMoves then self.index = 1; end -- Reset index to overwrite first element of table.
		self.members[ self.index ] = o;
		self.index = self.index + 1;
	end	
end
function m_UnitMoves:save( o ) -- Overwrites updated move object data into the table.
	if not o then print( "Error in m_UnitMoves:save( o ): Nil or invalid input" ); return; end
	for key, aMove in ipairs( self.members ) do 
		if ( aMove.ID == o.ID and ( ( aMove.owner == o.owner ) or ( aMove.owner == o.owner ) ) ) then self.members[key] = o; end
	end
end
function  m_UnitMoves:get( unit ) -- Returns an incomplete Unit Move object from the table.
	if not unit then print( "Error in m_UnitMoves:get( unit ): Nil or invalid input" ); return nil; end
	if ( type( unit ) == "table" ) then -- Check if a table was given.
		for key, aMove in ipairs( self.members ) do 
			if ( ( aMove.completed == false ) and ( aMove.ignored == false ) ) then -- Move has not been completed or ignored.
				if ( unit.GetID ~= nil and aMove.ID == unit:GetID() and unit.GetOwner ~= nil and aMove.owner == unit:GetOwner() ) 
					or ( aMove.ID == unit.ID and aMove.owner == unit.owner ) then 
					local shouldIgnore = aMove:ignore();
					if shouldIgnore == true then
						self.members[ key ] = aMove;
						return false;
					else
						return aMove;
					end
				end
			end
		end 
	end return nil;
end
function m_UnitMoves:refresh() -- Go through all unit moves and check if any incomplete moves should be ignored. 
	for key, aMove in ipairs( self.members ) do 
		if ( ( aMove.completed == false ) and ( aMove.ignored == false ) ) then -- Move has not been completed or ignored.
			local shouldIgnore = aMove:ignore();
			if shouldIgnore == true then
				self.members[ key ] = aMove;
			end
		end
	end	
end
function m_UnitMoves:create( unit, iPlot ) -- Creates a new move object, returns true if a new move was created.
	if unit and unit.GetID then
		local currentMove = self:get( unit );
		if currentMove == nil then -- Check if unit is already listed as moving in table.
			local plot = Map.GetPlotByIndex( iPlot );
			if plot then
				local move = 
					m_UnitMoveClass:new{ 
						ID = unit:GetID();
						owner = unit:GetOwner();
						toPlot = iPlot;
						timeStarted = GetSystemTime();
					};
				self:add( move );
				return true;
			else
				print( "Error in m_UnitMoves:create( unit, iPlot ): Invalid plot ID given in 'iPlot'." );
			end
		else
			if m_debug then print( "m_UnitMoves:create( unit, iPlot ): This unit is already flagged as moving, did not add new move." ); end
		end
	else
		print( "Error in m_UnitMoves:create( unit, iPlot ): Invalid unit object given in 'unit'." );	
	end
	return false;
end
function m_UnitMoves:finishMove( unit )	-- Marks a move as completed so it will not block a units moves.
	local wasIgnored = true;
	local move = self:get( unit ); -- Get an incomplete move object for this unit.
	if move then
		move.completed = true;
		self:save( move );
	end
end


-- Ranged targets.
local m_RangedTargetPlots = {
	members = {},
	index = -1,
	count = 0
};
function m_RangedTargetPlots:flush()
	self.members = {};
	self.count = 0;
	self.index = -1;	
end
-- Set a plot ID as currently selected if available
function m_RangedTargetPlots:SetSelected( ID )
	for index, iPlot in ipairs( self.members) do
		if ID == iPlot then
			self.index = index;	
		end
	end
end
-- Get currently selected plot id.
function m_RangedTargetPlots:Get(  )
	if self.count > 0 then
		return self.members[ self.index ];
	else
		return -1;
	end
end
-- Get a new table of available ranged targets for pUnit
function m_RangedTargetPlots:Refresh( pUnit ) 
	self:flush();
	local tResults = UnitManager.GetOperationTargets( pUnit, UnitOperationTypes.RANGE_ATTACK );
	local allPlots = tResults[ UnitOperationResults.PLOTS ];
	if ( allPlots ~= nil ) then
		for i, modifier in ipairs( tResults[ UnitOperationResults.MODIFIERS ] ) do
			if ( modifier == UnitOperationResults.MODIFIER_IS_TARGET ) then	
				table.insert( self.members, allPlots[i] );
			end
		end
	end
	self.count = table.count( self.members );
	if table.count( self.members ) > 0 then
		self.index = 1;
		return true;
	else
		self.index = -1;
		return false;
	end		
end
-- Get forward in available ranged target plots.
function m_RangedTargetPlots:GetNext(  )
	if self.count < 1 then
		return -1;
	elseif self.count > 1 then
		self.index = self.index + 1;
		if self.index > self.count then self.index = 1; end
	end
	return self.members[ self.index ];
end
-- Go backwards in available ranged target plots.
function m_RangedTargetPlots:GetPrev(  )
	if self.count < 1 then
		return -1;
	elseif self.count > 1 then
		self.index = self.index - 1;
		if self.index < 1 then self.index = self.count; end
	end
	return self.members[ self.index ];
end

--------------------------------------------------------------------------------------------------------
-- If the given value is above or below the max, rescale it between 0 and the max - 1 (unless non_zero_based parameter is true).
-- The scaling is done circularly, that is -1 --> max value - 1, that is the values wrap around.
-- If the non_zero_based parameter is set to true, then scaling will be done between 1 and max value.
local function rescale( value, max_value, non_zero_based )
	local new_value = value;
	if non_zero_based == nil or non_zero_based == false then -- scale between 0 and max_value - 1
		while (new_value < 0) do new_value = new_value + max_value; end
		while (new_value >= max_value) do new_value = new_value - max_value; end
	elseif non_zero_based == true then -- scale between 1 and max_value value
		while (new_value <= 0) do new_value = new_value + max_value; end
		while (new_value > max_value) do new_value = new_value - max_value; end
	else
		print( "ERROR in rescale function. Unrecognized value for parameter: 'non-zero-based'!" );
	end
	return new_value;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Returns true if the given key was double tapped.
local function WasKeyDoubleTapped( uiKey )
	if ( uiKey == m_LastKeyUp ) then -- Same key as last released.
		local dif = GetSystemTime() - m_LastKeyUpTime;
		if dif < m_DoubleTapTimer then -- time difference was less than double press threshold.
			return true;	
		end
	end
	return false;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Returns number of seconds since m_IsNumpadKeyDown was set to true.
local function GetTimeNumpadKeyHeldDown()
	if ( m_IsNumpadKeyDown == true ) and ( m_NumpadKeyTimePushed > -1 ) then
		local current_time = GetSystemTime();
		local dif = current_time - m_NumpadKeyTimePushed;
		if dif > 0 then return dif; else return 0; end
	else
		print( "Error in GetTimeNumpadKeyHeldDown: Attempted to call function while numpad key is not down." );
		return 0;
	end
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Push numpad key down if nothing is already being pushed down, and start a timer.
local function NumpadKeyDown( uiKey )
	if ( m_IsNumpadKeyDown == false ) then
		if m_debug then print( "NumpadKeyDown: " .. tostring( uiKey ) ); end
		m_IsNumpadKeyDown = true;
		m_NumpadKey = uiKey;
		m_NumpadKeyTimePushed = GetSystemTime();
	end	
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Reset numpad key, if it matches the key that was being held down. Return true if anything changed (action should be taken).
local function NumpadKeyUp( uiKey )
	if ( m_IsNumpadKeyDown == true ) and ( uiKey == m_NumpadKey ) then
		local timeHeld = GetTimeNumpadKeyHeldDown();
		m_IsNumpadKeyDown = false;
		m_NumpadKey = -1;
		m_NumpadKeyTimePushed = -1;
		if timeHeld < setting_KeyDownIgnoreTimer then -- Key held down for a short time, return true.
			if m_debug then print( "NumpadKeyUp: " .. tostring( uiKey ) .. " released. Attempting action." ); end
			return true;
		else -- Key was held down for a longer time (past threshold), return false.
			if m_debug then print( "NumpadKeyUp: " .. tostring( uiKey ) .. " after " .. timeHeld .. " seconds. No action taken." ); end
			return false;
		end
	end	
	return false;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Returns true, action if input key is used by bindings table.
local function IsInputNumpadKeyAction( uiKey )
	for action, binding in pairs( m_bindings ) do
		if binding == uiKey then
			return true, action;
		end
	end
	return false, nil;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Returns true, action if input key is used by double tap bindings table.
local function IsInputDoubleNumpadKeyAction( uiKey )
	for action, binding in pairs( m_double_bindings ) do
		if binding == uiKey then
			return true, action;
		end
	end
	return false, nil;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Returns true if unit can move to x, y
local function CanUnitMoveTo( pUnit, x, y )
	if ( not pUnit ) or ( not pUnit.GetID ) or ( not x ) or ( not y ) then print( "ERROR: Invalid arguments to function CanUnitMoveTo." ); return false; end
	local tParameters = {};
	tParameters[UnitOperationTypes.PARAM_X] = x;
	tParameters[UnitOperationTypes.PARAM_Y] = y;
	local canMove = UnitManager.CanStartOperation( pUnit, UNIT_OPS["UNITOPERATION_MOVE_TO"].HASH, nil, tParameters );
	if canMove == true then
		local turnsList		: table;
		local obstacles		: table;
		local pathPlots		: table = {};
		-- Get the path the unit would need to take to move to the next plot
		pathPlots, turnsList, obstacles = UnitManager.GetMoveToPath( pUnit, Map.GetPlotIndex( x, y ) );
		-- Do not command unit to move if it will queue an action for next turn (not enough movement left).
		if m_debug then print( "Number of plots in the path: ", table.count( pathPlots ) ); end
		local nTurns = turnsList[ 2 ]; -- number of turns required to get to next plot.
		if nTurns and ( nTurns > 1 ) then
			if m_debug then print( "Path would take more than one turn." ); end
			canMove = false;
		end
	end
	if m_debug then
		print( "CanUnitMoveTo: player (" .. pUnit:GetOwner() .. ") unit (" .. pUnit:GetID() .. ") to x=" .. x .. "; y=" .. y .. "? " .. tostring( canMove ) );
	end
	return canMove;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Returns true if plot at x, y contains an enemy move-attackable by unit.
local function PlotIsAttackableEnemy( pUnit, x, y )
	if ( not pUnit ) or ( not pUnit.GetID ) or ( not x ) or ( not y ) then print( "ERROR: Invalid arguments to function PlotIsAttackableEnemy." ); return false; end
	local tParameters = {};
	tParameters[UnitOperationTypes.PARAM_X] = x;
	tParameters[UnitOperationTypes.PARAM_Y] = y;
	tParameters[UnitOperationTypes.PARAM_MODIFIERS] = UnitOperationMoveModifiers.ATTACK;
	local canAttack = UnitManager.CanStartOperation( pUnit, UNIT_OPS["UNITOPERATION_MOVE_TO"].HASH, nil, tParameters );
	if canAttack == true then
		if ( pUnit:GetCombat() > 0 ) or ( pUnit:GetRangedCombat() > 0 ) or ( pUnit:GetReligiousStrength() > 0 ) then
			local units = Units.GetUnitsInPlot( Map.GetPlot( x, y ) );
			if not units then return false, nil; end -- return false if there is no unit.
			for i, unit in ipairs( units ) do
				if unit:GetOwner() ~= Game.GetLocalPlayer() then
					if ( unit:GetCombat() > 0 ) or ( unit:GetRangedCombat() > 0 ) then 
						return true, unit;
					elseif ( unit:GetReligiousStrength() > 0 ) then 
						return true, unit; -- Religious combat units.
					end
				end
			end
			-- If we got here, no military/religious unit, so the valid attack move must be capturing a support or civilian unit.
			for i, unit in ipairs( units ) do
				if unit:GetOwner() ~= Game.GetLocalPlayer() then
					return true, unit; -- just return the first unit found not owned by local player.
				end
			end
		end
	end
	return false, nil;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Returns true if plot at x, y contains an enemy range attackable by unit.
local function PlotIsRangeAttackableEnemy( pUnit, x, y )
	if ( not pUnit ) or ( not pUnit.GetID ) or ( not x ) or ( not y ) then print( "ERROR: Invalid arguments to function PlotIsRangeAttackableEnemy." ); return false; end
	if ( pUnit:GetRangedCombat() > 0 ) then
		local units = Units.GetUnitsInPlot( Map.GetPlot( x, y ) );
		if not units then return false, nil; end -- return false if there is no unit.
		for i, unit in ipairs( units ) do
			if unit:GetOwner() ~= Game.GetLocalPlayer() then
				if ( unit:GetCombat() > 0 ) or ( unit:GetRangedCombat() > 0 ) then 
					local tParameters = {};
					tParameters[UnitOperationTypes.PARAM_X] = x;
					tParameters[UnitOperationTypes.PARAM_Y] = y;
					tParameters[UnitOperationTypes.PARAM_MODIFIERS] = UnitOperationMoveModifiers.NONE;
					if UnitManager.CanStartOperation( pUnit, UNIT_OPS["UNITOPERATION_RANGE_ATTACK"].HASH, nil, tParameters) then
						return true, unit;
					end
				elseif unit:GetFormationClass() == "FORMATION_CLASS_SUPPORT" then -- Check if its a targetable support unit.
					local tParameters = {};
					tParameters[UnitOperationTypes.PARAM_X] = x;
					tParameters[UnitOperationTypes.PARAM_Y] = y;
					tParameters[UnitOperationTypes.PARAM_MODIFIERS] = UnitOperationMoveModifiers.NONE;
					if UnitManager.CanStartOperation( pUnit, UNIT_OPS["UNITOPERATION_RANGE_ATTACK"].HASH, nil, tParameters) then
						return true, unit;
					end
				end
			end
		end
		-- If we got here, no combat or support units that we can range attack. Just try to select an attackable unit not owned by local player.
		for i, unit in ipairs( units ) do
			if unit:GetOwner() ~= Game.GetLocalPlayer() then
				local tParameters = {};
				tParameters[UnitOperationTypes.PARAM_X] = x;
				tParameters[UnitOperationTypes.PARAM_Y] = y;
				tParameters[UnitOperationTypes.PARAM_MODIFIERS] = UnitOperationMoveModifiers.NONE;
				if UnitManager.CanStartOperation( pUnit, UNIT_OPS["UNITOPERATION_RANGE_ATTACK"].HASH, nil, tParameters) then
					return true, unit;
				end
			end
		end
	end
	return false, nil;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Hide combat information.
local function HideEnemyInfo( )
	if m_InspectedEnemy then
		LuaEvents.UnitFlagManager_PointerExited( m_InspectedEnemy:GetOwner(), m_InspectedEnemy:GetID() );
		m_InspectedEnemy = nil;
	end 
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Show combat information.
local function ShowEnemyInfo( enemyUnit )
	-- Fake mouse over unit flag call to make UnitPanel show combat info
	--if m_InspectedEnemy then HideEnemyInfo(); end
	m_InspectedEnemy = enemyUnit;
	LuaEvents.UnitFlagManager_PointerEntered( m_InspectedEnemy:GetOwner(), m_InspectedEnemy:GetID() ); 
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Enter ranged attack selection
local function EnterRangedAttackSelection(  )
	if m_IsSelectingRangeAttack == false then
		if m_SelectedUnit then
			-- Check if selected unit can perform ranged attack.
			if UnitManager.CanStartOperation( m_SelectedUnit, UNIT_OPS["UNITOPERATION_RANGE_ATTACK"].HASH, nil, false, false ) == true then 
				m_IsSelectingRangeAttack = true;
				UI.SetInterfaceMode( InterfaceModeTypes.RANGE_ATTACK );
				m_RangedTargetPlots:Refresh( m_SelectedUnit );
				m_FocusedTargetPlot = m_RangedTargetPlots:Get();
				if m_FocusedTargetPlot ~= -1 then
					local plot = Map.GetPlotByIndex( m_FocusedTargetPlot );
					if plot then
						UILens.FocusHex( LensLayers.ATTACK_RANGE, m_FocusedTargetPlot );
						local x = plot:GetX();
						local y = plot:GetY();
						local attackable = true;
						local enemy = nil;
						attackable, enemy = PlotIsAttackableEnemy( m_SelectedUnit, x, y );
						if enemy then
							ShowEnemyInfo( enemy );
						end
						return true;
					end
				end
			end
		end
	end
	return false;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Exit ranged attack selection
local function ExitRangedAttackSelection(  )
	if m_IsSelectingRangeAttack == true then
		if ( m_FocusedTargetPlot ~= -1 ) then UILens.UnFocusHex( LensLayers.ATTACK_RANGE, m_FocusedTargetPlot ); end
		m_IsSelectingRangeAttack = false;
		UI.SetInterfaceMode( InterfaceModeTypes.SELECTION );
		m_RangedTargetPlots:flush();
		m_FocusedTargetPlot = -1;
		HideEnemyInfo();
	end	
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Select next ranged attack
local function OnSelectNext(  )
	if m_IsSelectingRangeAttack == true then
		if m_SelectedUnit then
			local nextPlotID = m_RangedTargetPlots:GetNext();
			if ( nextPlotID ~= -1 ) and ( nextPlotID ~= m_FocusedTargetPlot ) then
				UILens.UnFocusHex( LensLayers.ATTACK_RANGE, m_FocusedTargetPlot );
				m_FocusedTargetPlot	= nextPlotID;
				local plot = Map.GetPlotByIndex( m_FocusedTargetPlot );
				if plot then
					UILens.FocusHex( LensLayers.ATTACK_RANGE, nextPlotID );
					local x = plot:GetX();
					local y = plot:GetY();
					local attackable = true;
					local enemy = nil;
					attackable, enemy = PlotIsAttackableEnemy( m_SelectedUnit, x, y );
					if enemy then
						ShowEnemyInfo( enemy );
					end
				end
			end
		end	
	end	
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Select previous ranged attack
local function OnSelectPrev(  )
	if m_IsSelectingRangeAttack == true then
		if m_SelectedUnit then
			local nextPlotID = m_RangedTargetPlots:GetPrev();
			if ( nextPlotID ~= -1 ) and ( nextPlotID ~= m_FocusedTargetPlot ) then
				UILens.UnFocusHex( LensLayers.ATTACK_RANGE, m_FocusedTargetPlot );
				m_FocusedTargetPlot	= nextPlotID;
				local plot = Map.GetPlotByIndex( m_FocusedTargetPlot );
				if plot then
					UILens.FocusHex( LensLayers.ATTACK_RANGE, nextPlotID );
					local x = plot:GetX();
					local y = plot:GetY();
					local attackable = true;
					local enemy = nil;
					attackable, enemy = PlotIsAttackableEnemy( m_SelectedUnit, x, y );
					if enemy then
						ShowEnemyInfo( enemy );
					end
				end
			end
		end	
	end	
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Commands pUnit to skip this turn.
local function SkipUnitTurn( pUnit )
	if ( not pUnit ) or ( not pUnit.GetID ) or ( not m_SelectedUnit ) then 
		print( "ERROR: Invalid arguments to function SkipUnitTurn." );
		return false;
	end
	local tParameters = {};
	tParameters[UnitOperationTypes.PARAM_X] = pUnit:GetX();
	tParameters[UnitOperationTypes.PARAM_Y] = pUnit:GetY();
	if UnitManager.CanStartOperation( pUnit, UNIT_OPS["UNITOPERATION_SKIP_TURN"].HASH, nil, tParameters ) then
		UnitManager.RequestOperation( pUnit, UNIT_OPS["UNITOPERATION_SKIP_TURN"].HASH, tParameters );
		return true;
	end
	return false;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Commands unit to move to x, y
local function MoveUnit( pUnit, x, y )
	if ( not pUnit ) or ( not pUnit.GetID ) or ( not x ) or ( not y ) or ( not m_SelectedUnit ) then 
		print( "ERROR: Invalid arguments to function MoveUnit." );
		return false;
	end
	local canMove = m_UnitMoves:create( pUnit, Map.GetPlotIndex( x, y ) ); -- Try to create new unit move object.
	if canMove then -- 
		local tParameters = {};
		tParameters[UnitOperationTypes.PARAM_X] = x;
		tParameters[UnitOperationTypes.PARAM_Y] = y;
		UnitManager.RequestOperation( pUnit, UNIT_OPS["UNITOPERATION_MOVE_TO"].HASH, tParameters );
		-- Request a ranged attack if unit can start that, as the move to operation request may have failed.
		tParameters[UnitOperationTypes.PARAM_MODIFIERS] = UnitOperationMoveModifiers.NONE;
		if UnitManager.CanStartOperation( pUnit, UNIT_OPS["UNITOPERATION_RANGE_ATTACK"].HASH, nil, tParameters) then
			UnitManager.RequestOperation( pUnit, UNIT_OPS["UNITOPERATION_RANGE_ATTACK"].HASH, tParameters);
			ExitRangedAttackSelection();
		end
		return true;
	end
	return false;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Commands unit to move to x, y to attack
local function MoveUnitToAttack( pUnit, x, y )
	if ( not pUnit ) or ( not pUnit.GetID ) or ( not x ) or ( not y ) or ( not m_SelectedUnit ) then 
		print( "ERROR: Invalid arguments to function MoveUnitToAttack." );
		return false;
	end
	local canMove = m_UnitMoves:create( pUnit, Map.GetPlotIndex( x, y ) ); -- Try to create new unit move object.
	if canMove then -- 
		local tParameters = {};
		tParameters[UnitOperationTypes.PARAM_X] = x;
		tParameters[UnitOperationTypes.PARAM_Y] = y;
		tParameters[UnitOperationTypes.PARAM_MODIFIERS] = UnitOperationMoveModifiers.ATTACK;
		UnitManager.RequestOperation( pUnit, UNIT_OPS["UNITOPERATION_MOVE_TO"].HASH, tParameters );
		-- Request a ranged attack if unit can start that, as the move to attack operation request may have failed.
		tParameters[UnitOperationTypes.PARAM_MODIFIERS] = UnitOperationMoveModifiers.NONE;
		if UnitManager.CanStartOperation( pUnit, UNIT_OPS["UNITOPERATION_RANGE_ATTACK"].HASH, nil, tParameters) then
			UnitManager.RequestOperation( pUnit, UNIT_OPS["UNITOPERATION_RANGE_ATTACK"].HASH, tParameters);
			ExitRangedAttackSelection();
		end
		return true;
	end
	return false;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Commands unit to attack unit in plot at x, y
local function UnitRangeAttack( pUnit, x_i, y_i )
	if ( not pUnit ) or ( not pUnit.GetID ) or ( not m_SelectedUnit ) then 
		print( "ERROR: Invalid arguments to function UnitRangeAttack." );
		return false;
	end
	local x = x_i;
	local y = y_i
	if not ( x and y ) then
		local plot = Map.GetPlotByIndex( m_FocusedTargetPlot );
		if plot and plot.GetX then
			x = plot:GetX();
			y = plot:GetY();
		else return false; end
	end	
	local tParameters = {};
	tParameters[UnitOperationTypes.PARAM_X] = x;
	tParameters[UnitOperationTypes.PARAM_Y] = y;
	tParameters[UnitOperationTypes.PARAM_MODIFIERS] = UnitOperationMoveModifiers.NONE;
	if UnitManager.CanStartOperation( pUnit, UNIT_OPS["UNITOPERATION_RANGE_ATTACK"].HASH, nil, tParameters) then
		UnitManager.RequestOperation( pUnit, UNIT_OPS["UNITOPERATION_RANGE_ATTACK"].HASH, tParameters);
		ExitRangedAttackSelection();
		return true;
	end
	return false;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Attempt to perform an action.
local function RealizeBindingAction( action:string, isDoubleTap:boolean )
	if not action then print( "ERROR: Invalid argument to function RealizeBindingAction." ); return false; end
	if m_SelectedUnit and m_SelectedUnit.GetID then
		local performMove = true;
		local doubleTap = isDoubleTap == true;
		local x_change = 0;
		local y_change = 0;
		local unit_y = m_SelectedUnit:GetY();
		local isYOdd = true;
		local modulus = unit_y - ( math.floor( unit_y / 2.0 ) * 2 );
		if modulus == 0 then isYOdd = false; end
		local direction = "";		   
		if action == "CENTER_CAMERA" then
			performMove = false;
			UI.LookAtPlot( Map.GetPlot( m_SelectedUnit:GetX(), m_SelectedUnit:GetY() ) );
		elseif action == "SELECT_NEXT" then
			performMove = false;
			UI.PlaySound("Play_UI_Click");
			if m_IsSelectingRangeAttack == true then
				OnSelectNext();
			else
				UI.SelectNextReadyUnit();
			end		
		elseif action == "SELECT_PREV" then
			performMove = false;
			UI.PlaySound("Play_UI_Click");
			if m_IsSelectingRangeAttack == true then
				OnSelectPrev();
			else
				UI.SelectPrevReadyUnit();
			end
		elseif action == "ATTACK_RANGE" then
			if doubleTap == true then
				performMove = false;
				if m_IsSelectingRangeAttack == true then -- Perform ranged attack
					UnitRangeAttack( m_SelectedUnit ); 
				else -- Enter ranged attack selection
					local canRangeAttack = EnterRangedAttackSelection();
					if canRangeAttack == false then
						SkipUnitTurn( m_SelectedUnit ); -- Double tap "5" will skip this unit's turn.
					end
				end
			end
		elseif action == "MOVE_SW" then
			if m_IsSelectingRangeAttack == true then -- Exit ranged attack selection.
				performMove = false;
				ExitRangedAttackSelection();
			else -- move
				direction = "South-West";
				if isYOdd == false then x_change = -1; else x_change = 0; end
				y_change = -1;	
			end
		elseif action == "MOVE_W" then -- Select prev range attack
			if m_IsSelectingRangeAttack == true then
				performMove = false;
				UI.PlaySound("Play_UI_Click");
				OnSelectPrev();
			else -- move
				direction = "West";
				x_change = -1;
				y_change = 0;
			end
		elseif action == "MOVE_NW" then
			if m_IsSelectingRangeAttack == true then -- Exit ranged attack selection.
				performMove = false;
				ExitRangedAttackSelection();
			else -- move
				direction = "North-West";
				if isYOdd == false then x_change = -1; else x_change = 0; end
				y_change = 1;	
			end
		elseif action == "MOVE_NE" then
			if m_IsSelectingRangeAttack == true then -- Exit ranged attack selection.
				performMove = false;
				ExitRangedAttackSelection();
			else -- move
				direction = "North-East";
				if isYOdd == false then x_change = 0; else x_change = 1; end
				y_change = 1;	
			end
		elseif action == "MOVE_E" then
			if m_IsSelectingRangeAttack == true then -- Select next range attack
				performMove = false;
				UI.PlaySound("Play_UI_Click");
				OnSelectNext();
			else -- move unit
				direction = "East";
				x_change = 1;
				y_change = 0;
			end
		elseif action == "MOVE_SE" then
			if m_IsSelectingRangeAttack == true then -- Exit ranged attack selection.
				performMove = false;
				ExitRangedAttackSelection();
			else -- move
				direction = "South-East";
				if isYOdd == false then x_change = 0; else x_change = 1; end
				y_change = -1;	
			end
		else
			return false;
		end
		if performMove == true then -- Try to move the unit.
			local previousMove = m_UnitMoves:get( m_SelectedUnit );
			if not previousMove then -- Don't do anything if a unit is currently moving.
				if ( x_change ~= 0 ) or ( y_change ~= 0	) then
					local x = m_SelectedUnit:GetX() + x_change;
					local y = m_SelectedUnit:GetY() + y_change;
					x = rescale( x, map_width, false ); -- Rescale x around based on map width.
					local canMove = CanUnitMoveTo( m_SelectedUnit, x, y );
					if canMove == true then	
						local attackable = false;
						local enemyUnit = nil;
						attackable, enemyUnit = PlotIsAttackableEnemy( m_SelectedUnit, x, y ); -- Can attack enemy combat unit?
						if attackable == true then
							if doubleTap == true then	-- Only proceed with attack on double tap.
								MoveUnitToAttack( m_SelectedUnit, x, y );
								if m_debug then print( "Key was double tapped. Moved unit (" .. m_SelectedUnit:GetID() .. ") to the " .. direction .. " for an attack." ); end	
							else
								ShowEnemyInfo( enemyUnit ); -- Show enemy combat info.
							end	
						else
							MoveUnit( m_SelectedUnit, x, y );
							if m_debug then print( "Moved unit (" .. m_SelectedUnit:GetID() .. ") to the " .. direction .. "." ); end
						end	
					else -- Can't move to the plot. Check if range attack available. 
						local attackable = false;
						local enemyUnit = nil;
						attackable, enemyUnit = PlotIsRangeAttackableEnemy( m_SelectedUnit, x, y );	-- Can range attack enemy combat unit?
						if attackable == true then
							if doubleTap == true then	-- Only proceed with attack on double tap.
								UnitRangeAttack( m_SelectedUnit, x, y );
								if m_debug then print( "Key was double tapped. Commanded unit (" .. m_SelectedUnit:GetID() .. 
									") to perform ranged attack to the " .. direction .. "." ); end	
							else
								ShowEnemyInfo( enemyUnit ); -- Show enemy combat info.
							end	
						end	
					end
				end
			end
		end
		return true;	
	end
	return false;
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Input Handler
function OnInputHandler( pInputStruct:table )

	local b_isHandled = false;
	if m_OkayToProcess == true then	-- Unit owned by local player is selected, okay to check keybinds.
		local uiMsg = pInputStruct:GetMessageType();
		
		if ( uiMsg == KeyEvents.KeyUp ) then -- User let off a key.
			local uiKey:number = pInputStruct:GetKey();
			local validKeyUp = NumpadKeyUp( uiKey );
			if validKeyUp == true then
				if uiKey == Keys.VK_ESCAPE then -- Escape key: get out of ranged selection
					if m_IsSelectingRangeAttack	== true then
						ExitRangedAttackSelection();
						return true;
					end
				elseif uiKey == Keys.VK_MULTIPLY then -- * key: get out of ranged selection
					if m_IsSelectingRangeAttack	== true then
						ExitRangedAttackSelection();
						return true;
					end
				end
				local doubleTapped = WasKeyDoubleTapped( uiKey );
				local action = nil;
				if doubleTapped == true then
					b_isHandled, action = IsInputDoubleNumpadKeyAction( uiKey );
				else
					b_isHandled, action = IsInputNumpadKeyAction( uiKey );
				end
				if action then -- Input is a binding.
					b_isHandled = RealizeBindingAction( tostring( action ), doubleTapped );
				end
			end
			m_LastKeyUp = uiKey; -- Mark this key as being the last pressed.
			m_LastKeyUpTime = GetSystemTime(); -- Save time when last key was released.

		elseif ( uiMsg == KeyEvents.KeyDown ) then -- User pressed key down.
			local uiKey:number = pInputStruct:GetKey();
			local action = nil;
			b_isHandled, action = IsInputNumpadKeyAction( uiKey );
			if action then -- Input is a binding.
				NumpadKeyDown( uiKey );
			else
				b_isHandled, action = IsInputDoubleNumpadKeyAction( uiKey );
				if action then -- Input is a binding.
					NumpadKeyDown( uiKey );	
				else
					if m_InspectedEnemy then -- Cancel showing unit info on any key down.
						HideEnemyInfo();
					end
				end
			end
		end
	else -- Unit is not selected. If a keybind was pressed, select next available unit.
		local uiMsg = pInputStruct:GetMessageType();
		if ( uiMsg == KeyEvents.KeyUp ) then -- User let off a key.
			local uiKey:number = pInputStruct:GetKey();
			local validKeyUp = NumpadKeyUp( uiKey );
			if validKeyUp == true then
				local doubleTapped = WasKeyDoubleTapped( uiKey );
				local action = nil;
				if doubleTapped == true then
					b_isHandled, action = IsInputDoubleNumpadKeyAction( uiKey );
				else
					b_isHandled, action = IsInputNumpadKeyAction( uiKey );
				end
				if action then -- Input is a binding.
					UI.SelectNextReadyUnit(); -- No unit selected, so select next available unit.
					UI.PlaySound("Play_UI_Click");
					b_isHandled = true;
				end
			end
			m_LastKeyUp = -1; -- Since a unit is not currently selected, don't allow this to be flagged as a double tap.
			m_LastKeyUpTime = -1; -- Save time when last key was released.

		elseif ( uiMsg == KeyEvents.KeyDown ) then -- User pressed key down.
			local uiKey:number = pInputStruct:GetKey();
			local action = nil;
			b_isHandled, action = IsInputNumpadKeyAction( uiKey );
			if action then -- Input is a binding.
				NumpadKeyDown( uiKey );
			else
				b_isHandled, action = IsInputDoubleNumpadKeyAction( uiKey );
				if action then -- Input is a binding.
					NumpadKeyDown( uiKey );	
				else
					if m_InspectedEnemy then -- Cancel showing unit info on any non numpad key down.
						HideEnemyInfo();
					end
				end
			end
		end
	end

	return b_isHandled; -- Return true if any input matches a keybind. Return false if not so it gets handled by other scripts.

end

-------------------------------------------------------------------------------------------------------------------------------------
-- Event handler: Set input handler flag and save selected unit data. 
local function OnUnitSelectionChanged( playerID, unitID, hexI, hexJ, hexK, isSelected, isEditable )
	if ( not Game ) or ( not Game.GetLocalPlayer ) or ( playerID ~= Game.GetLocalPlayer() ) then
		return;
	end
	if m_InspectedEnemy then -- Cancel showing unit info on any unit selection change.
		HideEnemyInfo();
	end
	if isSelected and ( not UI.IsGameCoreBusy() ) then
		local pSelectedUnit:table = UI.GetHeadSelectedUnit();
		if pSelectedUnit then
			m_OkayToProcess = true;
			m_SelectedUnit = pSelectedUnit;
		end
	else
		if m_IsSelectingRangeAttack == true then ExitRangedAttackSelection(); end -- Get out of range attack selection if unit is de-selected.
		m_OkayToProcess = false;
		m_SelectedUnit = nil;
		m_UnitMoves:refresh();	-- Check if any moves should be ignored and if so cancel pending unit operations.
	end
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Event handler: On unit move completion, free up new movement commands.
local function OnUnitMoveComplete( iPlayer, iUnit )
	if Game and Game.GetLocalPlayer and ( Game.GetLocalPlayer() == iPlayer ) then
		if m_InspectedEnemy then -- Cancel showing unit info on any unit move complete.
			HideEnemyInfo();
		end
		local unit = UnitManager.GetUnit( iPlayer, iUnit );
		if unit then
			m_UnitMoves:finishMove( unit );
		end	
	end
end

-------------------------------------------------------------------------------------------------------------------------------------
-- Run on script load.
local function Initialize()
	
	-- Enforce limits on timer values.
	if setting_KeyDownIgnoreTimer < 1 then setting_KeyDownIgnoreTimer = 1; end
	if setting_KeyDownIgnoreTimer > 6 then setting_KeyDownIgnoreTimer = 6; end

	ContextPtr:SetInputHandler( OnInputHandler, true );

	Events.UnitSelectionChanged.Add( OnUnitSelectionChanged );
	Events.UnitMoveComplete.Add( OnUnitMoveComplete );

end

Initialize();